"""Listener for converting ANTLR parse into Sentence expressions."""
from typing import Iterable, List, Optional

from antlr4 import CommonTokenStream, InputStream, ParseTreeWalker

from .expression import (
    ListReference,
    RuleReference,
    Sentence,
    Sequence,
    SequenceType,
    TextChunk,
)
from .grammar.HassILGrammarLexer import HassILGrammarLexer
from .grammar.HassILGrammarListener import HassILGrammarListener
from .grammar.HassILGrammarParser import HassILGrammarParser
from .util import normalize_text, remove_escapes

GROUP_MARKER = None


class HassILExpressionListener(HassILGrammarListener):
    """Listener that transforms ANTLR form into expressions."""

    def __init__(self) -> None:
        # List of sentences parsed so far
        self.sentences: List[Sentence] = []

        # Current sentence
        self._sentence: Optional[Sentence] = None

        # Current sequences being parsed.
        # The last sequence is the current target.
        # Groups of sequences and sub-sequences are separated with None.
        self._sequences: List[Optional[Sequence]] = []

        self._last_chunk: Optional[TextChunk] = None

    def enterSentence(self, ctx: HassILGrammarParser.SentenceContext):
        # Begin new sentence
        self._sentence = Sentence()
        self._sequences = [self._sentence]

    def exitSentence(self, ctx):
        # Complete sentence
        assert self._sentence is not None, "No sentence started"
        self.sentences.append(self._sentence)
        self._sentence = None
        self._sequences = []

    def enterGroup(self, ctx):
        # Begin new (group), not a sub-sequence
        self._sequences.append(GROUP_MARKER)
        self._sequences.append(Sequence(type=SequenceType.GROUP))

    def exitGroup(self, ctx):
        # Complete (group)
        group = self._pop_group()

        if (group.type == SequenceType.ALTERNATIVE) and (self._last_chunk is not None):
            self._last_chunk.text = self._last_chunk.text.rstrip()

        self.last_sequence.items.append(group)

    def enterOptional(self, ctx):
        # Begin new [optional], not a sub-sequence
        self._sequences.append(GROUP_MARKER)

        # Start as a group, will convert to alternative at exit
        self._sequences.append(Sequence(type=SequenceType.GROUP))

    def exitOptional(self, ctx):
        # Complete [optional], convert to alternative if necessary
        optional = self._pop_group()

        if optional.type != SequenceType.ALTERNATIVE:
            if len(optional.items) > 1:
                # Wrap in group
                optional.items = [
                    Sequence(type=SequenceType.GROUP, items=optional.items)
                ]

            optional.type = SequenceType.ALTERNATIVE

        if self._last_chunk is not None:
            self._last_chunk.text = self._last_chunk.text.rstrip()

        # Optionals are just alternatives with an empty text chunk ("") as an option.
        optional.items.append(TextChunk.empty())
        self.last_sequence.items.append(optional)

    def enterAlt(self, ctx):
        # Encountered an alternative marker "|" inside a group or optional.
        # This is complex because the group/optional may already contain some items.
        sequence = self.last_parent_sequence
        if sequence.type != SequenceType.ALTERNATIVE:
            # Convert to alternative
            if sequence.items:
                # Wrap into a group
                old_items = sequence.items
                sequence.items = [Sequence(type=SequenceType.GROUP, items=old_items)]

            sequence.type = SequenceType.ALTERNATIVE

        if self._last_chunk is not None:
            self._last_chunk.text = self._last_chunk.text.rstrip()

        # Start new group (sub-sequence)
        next_group = Sequence(type=SequenceType.GROUP)
        sequence.items.append(next_group)
        self._sequences.append(next_group)

    def enterRule(self, ctx: HassILGrammarParser.RuleContext):
        # <expansion_rule>
        rule_name = ctx.rule_name().STRING().getText()
        self.last_sequence.items.append(RuleReference(rule_name))

    def enterList(self, ctx: HassILGrammarParser.ListContext):
        # {slot_list}
        list_name = ctx.list_name().STRING().getText()
        self.last_sequence.items.append(ListReference(list_name))

    def enterText_chunk(self, ctx: HassILGrammarParser.Text_chunkContext):
        # Includes whitespace
        chunk_text = normalize_text(ctx.getText())
        chunk_text = remove_escapes(chunk_text)

        # Add to last sub-sequence
        self._last_chunk = TextChunk(chunk_text)
        self.last_sequence.items.append(self._last_chunk)

    def parse_sentences(self, sentences: Iterable[str]):
        """Parse multiple sentences separated by newlines."""
        text = "\n".join(sentences) + "\n"
        if text.strip():
            parser = HassILGrammarParser(
                CommonTokenStream(HassILGrammarLexer(InputStream(text)))
            )
            walker = ParseTreeWalker()
            walker.walk(self, parser.document())

    @property
    def last_sequence(self) -> Sequence:
        """Current sequence or sub-sequence target."""
        assert self._sequences, "No target sequence"
        last_sequence = self._sequences[-1]
        assert last_sequence, "Invalid target sequence"
        return last_sequence

    @property
    def last_parent_sequence(self) -> Sequence:
        """Current sequence target."""
        assert self._sequences, "No target sequence"

        # Locate last group marker to determine parent:
        # [parent, child, child, None, parent, ...]
        parent_sequence: Optional[Sequence] = None
        for sequence in reversed(self._sequences):
            if sequence is GROUP_MARKER:
                break

            parent_sequence = sequence

        assert parent_sequence, "No parent sequence"
        return parent_sequence

    def _pop_group(self) -> Sequence:
        """Remove sequences until a group marker is encountered."""
        assert len(self._sequences) > 1, "Bad sequence stack"
        item = self._sequences.pop()
        while self._sequences[-1] is not GROUP_MARKER:
            item = self._sequences.pop()

        assert item is not None, "Invalid last sequence"

        # Remove group marker
        marker = self._sequences.pop()
        assert marker is GROUP_MARKER, "Missing group marker"

        return item


def _trim_whitespace(seq: Sequence, first_chunk=None, last_chunk=None):
    pass



from hassil.intents import RangeSlotList, TextSlotList
from hassil.parse import parse_sentence
from hassil.sample import sample_expression


def test_text_chunk():
    assert set(sample_expression(parse_sentence("this is a test"))) == {
        "this is a test"
    }


def test_group():
    assert set(sample_expression(parse_sentence("this (is a) test"))) == {
        "this is a test"
    }


def test_optional():
    assert set(sample_expression(parse_sentence("turn on [the] light[s]"))) == {
        "turn on light",
        "turn on lights",
        "turn on the light",
        "turn on the lights",
    }


def test_alternative():
    assert set(sample_expression(parse_sentence("this is (the | a) test"))) == {
        "this is a test",
        "this is the test",
    }


def test_list():
    sentence = parse_sentence("turn off {area}")
    areas = TextSlotList.from_strings(["kitchen", "living room"])
    assert set(sample_expression(sentence, slot_lists={"area": areas})) == {
        "turn off kitchen",
        "turn off living room",
    }


def test_list_range():
    sentence = parse_sentence("run test {num}")
    num_list = RangeSlotList(1, 3)
    assert set(sample_expression(sentence, slot_lists={"num": num_list})) == {
        "run test 1",
        "run test 2",
        "run test 3",
    }


def test_rule():
    sentence = parse_sentence("turn off <area>")
    assert set(
        sample_expression(
            sentence,
            expansion_rules={"area": parse_sentence("[the] kitchen")},
        )
    ) == {"turn off kitchen", "turn off the kitchen"}



import io
from typing import cast

import pytest

from hassil import Intents, recognize, recognize_all
from hassil.expression import TextChunk
from hassil.intents import TextSlotList

TEST_YAML = """
language: "en"
intents:
  TurnOnTV:
    data:
      - sentences:
          - "turn on [the] TV in <area>"
          - "turn on <area> TV"
        slots:
          domain: "media_player"
          name: "roku"
  SetBrightness:
    data:
      - sentences:
          - "set [the] brightness in <area> to <brightness>"
        slots:
          domain: "light"
          name: "all"
      - sentences:
          - "set [the] brightness of <name> to <brightness>"
        requires_context:
          domain: "light"
        slots:
          domain: "light"
  GetTemperature:
    data:
      - sentences:
          - "<what_is> [the] temperature in <area>"
        slots:
          domain: "climate"
  CloseCover:
    data:
      - sentences:
          - "close <name>"
        requires_context:
          domain: "cover"
        slots:
          domain: "cover"
  Play:
    data:
      - sentences:
          - "play <name>"
        excludes_context:
          domain:
            - "cover"
            - "light"
  CloseCurtains:
    data:
      - sentences:
          - "close [the] curtains [in <area>]"
        slots:
          domain: "cover"
          device_class: "curtain"
        requires_context:
          area:
expansion_rules:
  area: "[the] {area}"
  name: "[the] {name}"
  brightness: "{brightness_pct} [percent]"
  what_is: "(what's | whats | what is)"
lists:
  brightness_pct:
    range:
      type: percentage
      from: 0
      to: 100
skip_words:
  - "please"
"""


@pytest.fixture
def intents():
    with io.StringIO(TEST_YAML) as test_file:
        return Intents.from_yaml(test_file)


@pytest.fixture
def slot_lists():
    return {
        "area": TextSlotList.from_tuples(
            [("kitchen", "area.kitchen"), ("living room", "area.living_room")]
        ),
        "name": TextSlotList.from_tuples(
            [
                ("hue", "light.hue", {"domain": "light"}),
                (
                    "garage door",
                    "cover.garage_door",
                    {"domain": "cover"},
                ),
                (
                    "blue curtains",
                    "cover.blue_curtains",
                    {
                        "domain": "cover",
                        "device_class": "curtain",
                        "area": "living_room",
                    },
                ),
                (
                    "roku",
                    "media_player.roku",
                    {"domain": "media_player"},
                ),
            ]
        ),
    }


# pylint: disable=redefined-outer-name
def test_turn_on(intents, slot_lists):
    result = recognize("turn on kitchen TV, please", intents, slot_lists=slot_lists)
    assert result is not None
    assert result.intent.name == "TurnOnTV"

    area = result.entities["area"]
    assert area.name == "area"
    assert area.value == "area.kitchen"

    # From YAML
    assert result.entities["domain"].value == "media_player"
    assert result.entities["name"].value == "roku"


# pylint: disable=redefined-outer-name
def test_brightness_area(intents, slot_lists):
    result = recognize(
        "set the brightness in the living room to 75%", intents, slot_lists=slot_lists
    )
    assert result is not None
    assert result.intent.name == "SetBrightness"

    assert result.entities["area"].value == "area.living_room"
    assert result.entities["brightness_pct"].value == 75

    # From YAML
    assert result.entities["domain"].value == "light"
    assert result.entities["name"].value == "all"


# pylint: disable=redefined-outer-name
def test_brightness_name(intents, slot_lists):
    result = recognize(
        "set brightness of the hue to 50%", intents, slot_lists=slot_lists
    )
    assert result is not None
    assert result.intent.name == "SetBrightness"

    assert result.entities["name"].value == "light.hue"
    assert result.entities["brightness_pct"].value == 50

    # From YAML
    assert result.entities["domain"].value == "light"


# pylint: disable=redefined-outer-name
def test_brightness_not_cover(intents, slot_lists):
    result = recognize(
        "set brightness of the garage door to 50%", intents, slot_lists=slot_lists
    )
    assert result is None


# pylint: disable=redefined-outer-name
def test_temperature(intents, slot_lists):
    result = recognize(
        "what is the temperature in the living room?", intents, slot_lists=slot_lists
    )
    assert result is not None
    assert result.intent.name == "GetTemperature"

    assert result.entities["area"].value == "area.living_room"

    # From YAML
    assert result.entities["domain"].value == "climate"


# pylint: disable=redefined-outer-name
def test_close_name(intents, slot_lists):
    result = recognize("close the garage door", intents, slot_lists=slot_lists)
    assert result is not None
    assert result.intent.name == "CloseCover"

    assert result.entities["name"].value == "cover.garage_door"

    # From YAML
    assert result.entities["domain"].value == "cover"


# pylint: disable=redefined-outer-name
def test_close_not_light(intents, slot_lists):
    result = recognize("close the hue", intents, slot_lists=slot_lists)
    assert result is None


# pylint: disable=redefined-outer-name
def test_play(intents, slot_lists):
    result = recognize("play roku", intents, slot_lists=slot_lists)
    assert result is not None
    assert result.intent.name == "Play"

    assert result.entities["name"].value == "media_player.roku"


# pylint: disable=redefined-outer-name
def test_play_no_cover(intents, slot_lists):
    result = recognize("play the garage door", intents, slot_lists=slot_lists)
    assert result is None


# pylint: disable=redefined-outer-name
def test_requires_context_implicit(intents, slot_lists):
    intent_context = {"area": "living room"}

    result = recognize(
        "close the curtains",
        intents,
        slot_lists=slot_lists,
        intent_context=intent_context,
    )
    assert result is not None
    assert result.intent.name == "CloseCurtains"

    assert result.entities["domain"].value == "cover"
    assert result.entities["device_class"].value == "curtain"


# pylint: disable=redefined-outer-name
def test_requires_context_none_provided(intents, slot_lists):
    result = recognize("close the curtains", intents, slot_lists=slot_lists)
    assert result is None


def test_lists_no_template() -> None:
    """Ensure list values without template syntax are plain text."""
    yaml_text = """
    language: "en"
    intents: {}
    lists:
      test:
        values:
          - "test value"
    """

    with io.StringIO(yaml_text) as test_file:
        intents = Intents.from_yaml(test_file)

    test_list = cast(TextSlotList, intents.slot_lists["test"])
    text_in = test_list.values[0].text_in
    assert isinstance(text_in, TextChunk)
    assert text_in.text == "test value"


def test_list_text_normalized() -> None:
    """Ensure list text in values are normalized."""
    yaml_text = """
    language: "en"
    intents:
      TestIntent:
        data:
          - sentences:
              - "run {test_name}"
    lists:
      test_name:
        values:
          - "tEsT    1"
    """

    with io.StringIO(yaml_text) as test_file:
        intents = Intents.from_yaml(test_file)

    result = recognize("run test 1", intents)
    assert result is not None
    assert result.entities["test_name"].value == "tEsT    1"


def test_skip_prefix() -> None:
    yaml_text = """
    language: "en"
    intents:
      TestIntent:
        data:
          - sentences:
              - "run {test_name}"
    lists:
      test_name:
        values:
          - "test"
    skip_words:
      - "the"
    """

    with io.StringIO(yaml_text) as test_file:
        intents = Intents.from_yaml(test_file)

    result = recognize("run the test", intents)
    assert result is not None
    assert result.entities["test_name"].value == "test"


def test_skip_sorted() -> None:
    """Ensure skip words are processed longest first"""
    yaml_text = """
    language: "en"
    intents:
      TestIntent:
        data:
          - sentences:
              - "run {test_name}"
    lists:
      test_name:
        values:
          - "test"
    skip_words:
      - "could"
      - "could you"
    """

    with io.StringIO(yaml_text) as test_file:
        intents = Intents.from_yaml(test_file)

    result = recognize("could you run test", intents)
    assert result is not None
    assert result.entities["test_name"].value == "test"


def test_response_key() -> None:
    """Check response key in intent data"""
    yaml_text = """
    language: "en"
    intents:
      TestIntent:
        data:
          - sentences:
              - "this is a test"
            response: "test_response"
    """

    with io.StringIO(yaml_text) as test_file:
        intents = Intents.from_yaml(test_file)

    result = recognize("this is a test", intents)
    assert result is not None
    assert result.response == "test_response"


def test_entity_text() -> None:
    """Ensure original text is returned as well as substituted list value"""
    yaml_text = """
    language: "en"
    intents:
      TestIntent:
        data:
          - sentences:
              - "run test {name} [now]"
              - "{name} test"
    lists:
      name:
        values:
          - in: "alpha"
            out: "A"
    """

    with io.StringIO(yaml_text) as test_file:
        intents = Intents.from_yaml(test_file)

    for sentence in ("run test alpha now", "run test alpha", "alpha test"):
        result = recognize(sentence, intents)
        assert result is not None
        assert result.entities["name"].value == "A"
        assert result.entities["name"].text.strip() == "alpha"


def test_number_text() -> None:
    """Ensure original text is returned as well as substituted number"""
    yaml_text = """
    language: "en"
    intents:
      TestIntent:
        data:
          - sentences:
              - "set {percentage} [now]"
              - "{percentage} set"
    lists:
      percentage:
        range:
          from: 0
          to: 100
    """

    with io.StringIO(yaml_text) as test_file:
        intents = Intents.from_yaml(test_file)

    for sentence in ("set 50% now", "set 50%", "50% set"):
        result = recognize(sentence, intents)
        assert result is not None
        assert result.entities["percentage"].value == 50
        assert result.entities["percentage"].text.strip() == "50%"


def test_recognize_all() -> None:
    """Test recognize_all method for returning all matches."""
    yaml_text = """
    language: "en"
    intents:
      TestIntent1:
        data:
          - sentences:
              - "run test"
      TestIntent2:
        data:
          - sentences:
              - "run test"
    """

    with io.StringIO(yaml_text) as test_file:
        intents = Intents.from_yaml(test_file)

    results = list(recognize_all("run test", intents))
    assert len(results) == 2
    assert {result.intent.name for result in results} == {
        "TestIntent1",
        "TestIntent2",
    }



"""Methods for recognizing intents from text."""

import collections.abc
import itertools
import re
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Optional

from .expression import (
    Expression,
    ListReference,
    RuleReference,
    Sentence,
    Sequence,
    SequenceType,
    TextChunk,
)
from .intents import Intent, Intents, RangeSlotList, SlotList, TextSlotList
from .util import normalize_text, normalize_whitespace

NUMBER_START = re.compile(r"^(\s*-?[0-9]+)")

NON_WORD_START = re.compile(r"^\W+")


class HassilError(Exception):
    """Base class for hassil errors"""


class MissingListError(HassilError):
    """Error when a {slot_list} is missing."""


class MissingRuleError(HassilError):
    """Error when an <expansion_rule> is missing."""


@dataclass
class MatchEntity:
    """Named entity that has been matched from a {slot_list}"""

    name: str
    """Name of the entity."""

    value: Any
    """Value of the entity."""

    text: str
    """Original value text."""


@dataclass
class MatchContext:
    """Context passed to match_expression."""

    text: str
    """Input text remaining to be processed."""

    slot_lists: Dict[str, SlotList] = field(default_factory=dict)
    """Available slot lists mapped by name."""

    expansion_rules: Dict[str, Sentence] = field(default_factory=dict)
    """Available expansion rules mapped by name."""

    entities: List[MatchEntity] = field(default_factory=list)
    """Entities that have been found in input text."""

    intent_context: Dict[str, Any] = field(default_factory=dict)
    """Context items from outside or acquired during matching."""

    @property
    def is_match(self) -> bool:
        """True if no text is left that isn't just whitespace or non-word characters"""
        text = re.sub(r"\W+", "", self.text).strip()
        return not text


@dataclass
class RecognizeResult:
    """Result of recognition."""

    intent: Intent
    """Matched intent"""

    entities: Dict[str, MatchEntity] = field(default_factory=dict)
    """Matched entities mapped by name."""

    entities_list: List[MatchEntity] = field(default_factory=list)
    """Matched entities as a list (duplicates allowed)."""

    response: Optional[str] = None
    """Key for intent response."""


def recognize(
    text: str,
    intents: Intents,
    slot_lists: Optional[Dict[str, SlotList]] = None,
    expansion_rules: Optional[Dict[str, Sentence]] = None,
    skip_words: Optional[Iterable[str]] = None,
    intent_context: Optional[Dict[str, Any]] = None,
    default_response: Optional[str] = "default",
) -> Optional[RecognizeResult]:
    """Return the first match of input text/words against a collection of intents."""
    for result in recognize_all(
        text,
        intents,
        slot_lists=slot_lists,
        expansion_rules=expansion_rules,
        skip_words=skip_words,
        intent_context=intent_context,
        default_response=default_response,
    ):
        return result

    return None


def recognize_all(
    text: str,
    intents: Intents,
    slot_lists: Optional[Dict[str, SlotList]] = None,
    expansion_rules: Optional[Dict[str, Sentence]] = None,
    skip_words: Optional[Iterable[str]] = None,
    intent_context: Optional[Dict[str, Any]] = None,
    default_response: Optional[str] = "default",
) -> Iterable[RecognizeResult]:
    """Return all matches for input text/words against a collection of intents."""
    text = normalize_text(text)

    if skip_words is None:
        skip_words = intents.skip_words
    else:
        # Combine skip words
        skip_words = itertools.chain(skip_words, intents.skip_words)

    if skip_words:
        text = _remove_skip_words(text, skip_words)

    text += " "

    if slot_lists is None:
        slot_lists = intents.slot_lists
    else:
        # Combine with intents
        slot_lists = {**intents.slot_lists, **slot_lists}

    if slot_lists is None:
        slot_lists = {}

    if expansion_rules is None:
        expansion_rules = intents.expansion_rules
    else:
        # Combine rules
        expansion_rules = {**intents.expansion_rules, **expansion_rules}

    if intent_context is None:
        intent_context = {}

    # Check sentence against each intent.
    # This should eventually be done in parallel.
    for intent in intents.intents.values():
        for intent_data in intent.data:
            for intent_sentence in intent_data.sentences:
                # Create initial context
                match_context = MatchContext(
                    text=text,
                    slot_lists=slot_lists,
                    expansion_rules=expansion_rules,
                    intent_context=intent_context,
                )
                maybe_match_contexts = match_expression(match_context, intent_sentence)
                for maybe_match_context in maybe_match_contexts:
                    if maybe_match_context.is_match:
                        skip_match = False

                        # Verify excluded context
                        if intent_data.excludes_context:
                            for (
                                context_key,
                                context_value,
                            ) in intent_data.excludes_context.items():
                                actual_value = maybe_match_context.intent_context.get(
                                    context_key
                                )
                                if actual_value == context_value:
                                    # Exact match to context value
                                    skip_match = True
                                    break

                                if (
                                    isinstance(
                                        context_value, collections.abc.Collection
                                    )
                                    and not isinstance(context_value, str)
                                    and (actual_value in context_value)
                                ):
                                    # Actual value was in context value list
                                    skip_match = True
                                    break

                        # Verify required context
                        if (not skip_match) and intent_data.requires_context:
                            for (
                                context_key,
                                context_value,
                            ) in intent_data.requires_context.items():
                                actual_value = maybe_match_context.intent_context.get(
                                    context_key
                                )

                                if (
                                    actual_value == context_value
                                    and context_value is not None
                                ):
                                    # Exact match to context value, except when context value is required and not provided
                                    continue

                                if context_value is None and actual_value is not None:
                                    # Any value matches, as long as it's set
                                    continue

                                if (
                                    isinstance(
                                        context_value, collections.abc.Collection
                                    )
                                    and not isinstance(context_value, str)
                                    and (actual_value in context_value)
                                ):
                                    # Actual value was in context value list
                                    continue

                                # Did not match required context
                                skip_match = True
                                break

                        if skip_match:
                            # Intent context did not match
                            continue

                        # Add fixed entities
                        for slot_name, slot_value in intent_data.slots.items():
                            maybe_match_context.entities.append(
                                MatchEntity(name=slot_name, value=slot_value, text="")
                            )

                        # Return the first match
                        response = default_response
                        if intent_data.response is not None:
                            response = intent_data.response

                        yield RecognizeResult(
                            intent=intent,
                            entities={
                                entity.name: entity
                                for entity in maybe_match_context.entities
                            },
                            entities_list=maybe_match_context.entities,
                            response=response,
                        )


def is_match(
    text: str,
    sentence: Sentence,
    slot_lists: Optional[Dict[str, SlotList]] = None,
    expansion_rules: Optional[Dict[str, Sentence]] = None,
    skip_words: Optional[Iterable[str]] = None,
    entities: Optional[Dict[str, Any]] = None,
    intent_context: Optional[Dict[str, Any]] = None,
) -> Optional[MatchContext]:
    """Return the first match of input text/words against a sentence expression."""
    text = normalize_text(text)

    if skip_words:
        text = _remove_skip_words(text, skip_words)

    text = text + " "

    if slot_lists is None:
        slot_lists = {}

    if expansion_rules is None:
        expansion_rules = {}

    if intent_context is None:
        intent_context = {}

    match_context = MatchContext(
        text=text,
        slot_lists=slot_lists,
        expansion_rules=expansion_rules,
        intent_context=intent_context,
    )

    for maybe_match_context in match_expression(match_context, sentence):
        if maybe_match_context.is_match:
            return maybe_match_context

    return None


def _remove_skip_words(text: str, skip_words: Iterable[str]) -> str:
    """Remove skip words from text."""

    # It's critical that skip words are processed longest first, since they may
    # share prefixes.
    for skip_word in sorted(skip_words, key=len, reverse=True):
        skip_word = normalize_text(skip_word)
        text = re.sub(rf"\b{re.escape(skip_word)}\b", "", text)

    text = normalize_whitespace(text)
    text = text.strip()

    return text


def match_expression(
    context: MatchContext, expression: Expression
) -> Iterable[MatchContext]:
    """Yield matching contexts for an expresion"""
    if isinstance(expression, TextChunk):
        chunk: TextChunk = expression
        chunk_text = chunk.text.lstrip()

        if chunk.is_empty:
            # Skip empty chunk
            yield context
        elif context.text.startswith(chunk_text):
            # Successful match for chunk
            context_text = context.text[len(chunk_text) :]
            context_text = context_text.lstrip()
            yield MatchContext(
                text=context_text,
                # Copy over
                slot_lists=context.slot_lists,
                expansion_rules=context.expansion_rules,
                entities=context.entities,
                intent_context=context.intent_context,
            )
        else:
            # Remove non-word characters and try again
            match = NON_WORD_START.match(context.text)
            if match is not None:
                context_text = context.text[len(match[0]) :].lstrip()
                if context_text.startswith(chunk_text):
                    context_text = context_text[len(chunk_text) :]
                    context_text = context_text.lstrip()
                    yield MatchContext(
                        text=context_text,
                        # Copy over
                        slot_lists=context.slot_lists,
                        expansion_rules=context.expansion_rules,
                        entities=context.entities,
                        intent_context=context.intent_context,
                    )

    elif isinstance(expression, Sequence):
        seq: Sequence = expression
        if seq.type == SequenceType.ALTERNATIVE:
            # Any may match (words | in | alternative)
            # NOTE: [optional] = (optional | )
            for item in seq.items:
                yield from match_expression(context, item)

        elif seq.type == SequenceType.GROUP:
            # All must match (words in group)
            if seq.items:
                group_contexts = [context]
                for item in seq.items:
                    # Next step
                    group_contexts = [
                        item_context
                        for group_context in group_contexts
                        for item_context in match_expression(group_context, item)
                    ]
                    if not group_contexts:
                        break

                for group_context in group_contexts:
                    yield group_context
        else:
            raise ValueError(f"Unexpected sequence type: {seq}")

    elif isinstance(expression, ListReference):
        # {list}
        list_ref: ListReference = expression
        if (not context.slot_lists) or (list_ref.list_name not in context.slot_lists):
            raise MissingListError(f"Missing slot list {{{list_ref.list_name}}}")

        slot_list = context.slot_lists[list_ref.list_name]
        if isinstance(slot_list, TextSlotList):
            text_list: TextSlotList = slot_list

            if context.text:
                # Any value may match
                for slot_value in text_list.values:
                    value_contexts = match_expression(
                        MatchContext(
                            # Copy over
                            text=context.text,
                            slot_lists=context.slot_lists,
                            expansion_rules=context.expansion_rules,
                            entities=context.entities,
                            intent_context=context.intent_context,
                        ),
                        slot_value.text_in,
                    )

                    for value_context in value_contexts:
                        entities = context.entities + [
                            MatchEntity(
                                name=list_ref.slot_name,
                                value=slot_value.value_out,
                                text=context.text[: -len(value_context.text)]
                                if value_context.text
                                else context.text,
                            )
                        ]

                        if slot_value.context:
                            # Merge context from matched list value
                            yield MatchContext(
                                entities=entities,
                                intent_context={
                                    **context.intent_context,
                                    **slot_value.context,
                                },
                                # Copy over
                                text=value_context.text,
                                slot_lists=value_context.slot_lists,
                                expansion_rules=value_context.expansion_rules,
                            )
                        else:
                            yield MatchContext(
                                entities=entities,
                                # Copy over
                                text=value_context.text,
                                slot_lists=value_context.slot_lists,
                                expansion_rules=value_context.expansion_rules,
                                intent_context=value_context.intent_context,
                            )

        elif isinstance(slot_list, RangeSlotList):
            # List that represents a number range.
            # Numbers must currently be digits ("1" not "one").
            range_list: RangeSlotList = slot_list
            if context.text:
                number_match = NUMBER_START.match(context.text)
                if number_match is not None:
                    number_text = number_match[1]
                    word_number = int(number_text)
                    if range_list.step == 1:
                        # Unit step
                        in_range = range_list.start <= word_number <= range_list.stop
                    else:
                        # Non-unit step
                        in_range = word_number in range(
                            range_list.start, range_list.stop + 1, range_list.step
                        )

                    if in_range:
                        entities = context.entities + [
                            MatchEntity(
                                name=list_ref.slot_name,
                                value=word_number,
                                text=context.text.split()[0],
                            )
                        ]

                        yield MatchContext(
                            text=context.text[len(number_text) :].lstrip(),
                            entities=entities,
                            # Copy over
                            slot_lists=context.slot_lists,
                            expansion_rules=context.expansion_rules,
                            intent_context=context.intent_context,
                        )

        else:
            raise ValueError(f"Unexpected slot list type: {slot_list}")

    elif isinstance(expression, RuleReference):
        # <rule>
        rule_ref: RuleReference = expression
        if (not context.expansion_rules) or (
            rule_ref.rule_name not in context.expansion_rules
        ):
            raise MissingRuleError(f"Missing expansion rule <{rule_ref.rule_name}>")

        yield from match_expression(
            context, context.expansion_rules[rule_ref.rule_name]
        )
    else:
        raise ValueError(f"Unexpected expression: {expression}")



from hassil.util import (
    is_template,
    merge_dict,
    normalize_text,
    normalize_whitespace,
    remove_escapes,
)


def test_merge_dict():
    base_dict = {"a": 1, "list": [1], "dict": {"a": 1}}
    merge_dict(base_dict, {"a": 2, "list": [2], "dict": {"b": 2}})
    assert base_dict == {"a": 2, "list": [1, 2], "dict": {"a": 1, "b": 2}}


def test_remove_escapes():
    assert remove_escapes("\\[test\\]") == "[test]"


def test_normalize_whitespace():
    assert normalize_whitespace("this    is a      test") == "this is a test"


def test_normalize_text():
    assert normalize_text("tHIS    is A      Test") == "this is a test"


def test_is_template():
    assert not is_template("just some plain text")
    assert is_template("[optional] word")
    assert is_template("a {list}")
    assert is_template("a <rule>")
    assert is_template("(a group)")
    assert is_template("an | alternative")



"""Convenience methods for parsing sentences."""

from typing import Iterable, List

from .expression import Sentence
from .expression_listener import HassILExpressionListener


def parse_sentences(texts: Iterable[str], keep_text: bool = False) -> List[Sentence]:
    """Parses a list of sentences into expressions."""
    listener = HassILExpressionListener()
    listener.parse_sentences(texts)
    if keep_text:
        for text, sentence in zip(texts, listener.sentences):
            sentence.text = text

    return listener.sentences


def parse_sentence(text: str, keep_text: bool = False) -> Sentence:
    """Parses a single sentence."""
    sentence = parse_sentences([text])[0]
    if keep_text:
        sentence.text = text

    return sentence
